[PATCH 4/7] percpu: enforce global uniqueness and disallow in-function statics

From: Tejun Heo
Date: Wed May 20 2009 - 03:41:46 EST


Some archs (alpha and s390) need to add 'weak' attribute to percpu
variable definitions so that the compiler generates external
references for them. To allow this, all percpu variables should be
unique regardless of scope and in-function static definitions can't be
allowed.

This patch updates DECLARE_PER_CPU_SECTION() and
DEFINE_PER_CPU_SECTION() such that percpu symbol itself is always
defined global so that weak can be added via PER_CPU_ATTRIBUTES. The
scope and uniqueness rules are enforced using two dummy variables,
both of which are thrown away during linking. The multiline
definition combined with section specification prevents in-function
static definitions.

This patch is inspired by Ivan Kokshaysky's alpha percpu patch.

[ Impact: stricter rules for percpu variable definitions ]

Signed-off-by: Tejun Heo <tj@xxxxxxxxxx>
Cc: Ingo Molnar <mingo@xxxxxxx>
Cc: David Howells <dhowells@xxxxxxxxxx>
Cc: Ivan Kokshaysky <ink@xxxxxxxxxxxxxxxxxxxx>
---
include/linux/percpu-defs.h | 34 +++++++++++++++++++++++++---------
1 files changed, 25 insertions(+), 9 deletions(-)

diff --git a/include/linux/percpu-defs.h b/include/linux/percpu-defs.h
index 8f921d7..6ff4fc1 100644
--- a/include/linux/percpu-defs.h
+++ b/include/linux/percpu-defs.h
@@ -10,21 +10,37 @@
/*
* Base implementations of per-CPU variable declarations and definitions, where
* the section in which the variable is to be placed is provided by the
* 'section' argument. This may be used to affect the parameters governing the
* variable's storage.
*
+ * Percpu variable declaration and definition decouple scope (static or extern)
+ * specification from the actual percpu variable and use two dummy variables
+ * to ensure DECLARE_PER_CPU_SECTION() is not coupled with static definition
+ * and each percpu symbol is unique regardless of scope. This is for archs
+ * which need to add 'weak' attribute to per cpu variables to force the
+ * compiler to generate external references for percpu variables (alpha and
+ * s390). This also means that in-function static percpu variables aren't
+ * allowed.
+ *
* NOTE! The sections for the DECLARE and for the DEFINE must match, lest
* linkage errors occur due the compiler generating the wrong code to access
* that section.
*/
-#define DECLARE_PER_CPU_SECTION(type, name, section) \
- extern \
- __attribute__((__section__(PER_CPU_BASE_SECTION section))) \
- PER_CPU_ATTRIBUTES __typeof__(type) per_cpu__##name
-
-#define DEFINE_PER_CPU_SECTION(type, name, section) \
- __attribute__((__section__(PER_CPU_BASE_SECTION section))) \
- PER_CPU_ATTRIBUTES __typeof__(type) per_cpu__##name
+#define __PCPU_ATTRS(sec) \
+ __attribute__((section(PER_CPU_BASE_SECTION sec))) \
+ PER_CPU_ATTRIBUTES
+
+#define __PCPU_DUMMY_ATTRS \
+ __attribute__((section(".discard"), unused))
+
+#define DECLARE_PER_CPU_SECTION(type, name, sec) \
+ extern __PCPU_DUMMY_ATTRS char __pcpu_scope_##name; \
+ extern __PCPU_ATTRS(sec) __typeof__(type) per_cpu__##name
+
+#define DEFINE_PER_CPU_SECTION(type, name, sec) \
+ __PCPU_DUMMY_ATTRS char __pcpu_scope_##name; \
+ __PCPU_DUMMY_ATTRS char __pcpu_unique_##name; \
+ __PCPU_ATTRS(sec) __typeof__(type) per_cpu__##name

/*
* Variant on the per-CPU variable declaration/definition theme used for
@@ -10,21 +10,37 @@
* Base implementations of per-CPU variable declarations and definitions, where
* the section in which the variable is to be placed is provided by the
- * 'section' argument. This may be used to affect the parameters governing the
+ * 'sec' argument. This may be used to affect the parameters governing the
* variable's storage.
*
--
1.6.0.2

--
To unsubscribe from this list: send the line "unsubscribe linux-kernel" in
the body of a message to majordomo@xxxxxxxxxxxxxxx
More majordomo info at http://vger.kernel.org/majordomo-info.html
Please read the FAQ at http://www.tux.org/lkml/